/*********************************************************************
 * Desktop Station Gateway Library for Raspberry Pi
 *
 * Copyright (C) 2015 Yaasan
 *
 */


#include <stdint.h>
#include <stdio.h>
#include <wiringPiSPI.h>
#include <unistd.h>
#include "DSGatewayLibPi.h"

DSGatewayLib::DSGatewayLib()
{

  //pinMode(SPI_SSPIN, OUTPUT);
  //digitalWrite(SPI_SSPIN, HIGH);
	
	poweron = false;
	
}

void DSGatewayLib::begin()
{
  /* Open SPI */
	if(wiringPiSPISetup(SPI_CHANNEL, SPI_SPEED) == -1){
		printf("could not initialize SPI\n");
		return;
	}
	//ioctl(

}

DSGatewayLib::~DSGatewayLib()
{
  
  SetPower(false);


}

bool DSGatewayLib::IsPower()
{
	
	return poweron;
	
}

void DSGatewayLib::clearMessage(unsigned char *inPackets)
{
 	for( int i = 0; i < SIZE_PACKET; i++)
	{
		inPackets[i] = 0;
	} 
}

bool DSGatewayLib::sendMessage(unsigned char *inPackets)
{
	uint8_t aReceived[SIZE_PACKET] = {0,0,0,0,0,0,0,0};
	int i;

	//digitalWrite(SPI_SSPIN, LOW);

	for( i = 0; i < SIZE_PACKET; i++)
	{
		aReceived[i] = reverseByte(inPackets[i]);
	}

	wiringPiSPIDataRW(SPI_CHANNEL, aReceived, SIZE_PACKET);

	//digitalWrite(SPI_SSPIN, HIGH);
	
	
	/* x */
	delay(30);
	
	/* Check for receiving */
	if( ((aReceived[1] & 0b11110000) == CMD_OK) ||((aReceived[1] & 0b11110000) == CMD_WAIT) )
	{
		return true;
	}
	else
	{
		return false;
	}
}

bool DSGatewayLib::exchangeMessage(unsigned char *inPackets, uint16_t timeout)
{
	//unsigned char aTemp[SIZE_PACKET];

	bool aReturnOk = sendMessage(inPackets);
	
	/*
	unsigned long aTime = millis();

	// response;

	while ((millis() < aTime + timeout) || (!aReturnOk)) {
		aTemp[0] = CMD_WAIT | 0b0010;
		aTemp[1] = aTemp[0];
		
		delay(20);
		
		if (sendMessage(aTemp) == true) {
			return true;
		}
	}

	if (DEBUG && !aReturnOk) {
		Serial.println(F("!!! Communication Error(Timeout, Command etc)"));
	}
	*/
	
	return aReturnOk;
}


bool DSGatewayLib::SetPower(bool power)
{
	unsigned char aPacktes[SIZE_PACKET];

	clearMessage(aPacktes);

	if (power) {
		aPacktes[0] = CMD_PWR_ON | 0b0010;
		aPacktes[1] = aPacktes[0];//CRC
		poweron = true;

	}
	else
	{
		aPacktes[0] = CMD_PWR_OFF | 0b0010;
		aPacktes[1] = aPacktes[0];//CRC
		poweron = false;
	}

	return exchangeMessage(aPacktes, TIME_REPLY);
}

bool DSGatewayLib::SetPowerEx(bool power)
{
	unsigned char aPacktes[SIZE_PACKET];

	clearMessage(aPacktes);

	if (power) {
		aPacktes[0] = CMD_PWR_ON | 0b0011;
		aPacktes[1] = 0b00000011;//Extend command (DCC128 and MM2-28step)
		aPacktes[2] = generateCRC(aPacktes, 2);

	}
	else
	{
		aPacktes[0] = CMD_PWR_OFF | 0b0010;
		aPacktes[1] = aPacktes[0];//CRC
	}
	


	return exchangeMessage(aPacktes, TIME_REPLY);
}

bool DSGatewayLib::SetLocoSpeed(uint16_t address, int inSpeed)
{
	unsigned char aPacktes[SIZE_PACKET];

	clearMessage(aPacktes);

	aPacktes[0] = CMD_SPEED | 0b0110;
	aPacktes[1] = lowbyte(address);
	aPacktes[2] = highbyte(address);
	aPacktes[3] = lowbyte(inSpeed);
	aPacktes[4] = highbyte(inSpeed);
	aPacktes[5] = generateCRC(aPacktes, 5);

	return exchangeMessage(aPacktes, TIME_REPLY);
}


bool DSGatewayLib::SetLocoSpeedEx(uint16_t address, int inSpeed, int inProtcol)
{
    unsigned char aPacktes[SIZE_PACKET];

    clearMessage(aPacktes);

    aPacktes[0] = CMD_SPEED | 0b0111;
    aPacktes[1] = lowbyte(address);
    aPacktes[2] = highbyte(address);
    aPacktes[3] = lowbyte(inSpeed);
    aPacktes[4] = highbyte(inSpeed);
    aPacktes[5] = inProtcol;
    aPacktes[6] = generateCRC(aPacktes, 6);

    return exchangeMessage(aPacktes, TIME_REPLY);
} 

bool DSGatewayLib::SetLocoFunction(uint16_t address, unsigned char inFunction, unsigned char inPower)
{
	unsigned char aPacktes[SIZE_PACKET];

	clearMessage(aPacktes);

	aPacktes[0] = CMD_FUNCTION | 0b0110;
	aPacktes[1] = lowbyte(address);
	aPacktes[2] = highbyte(address);
	aPacktes[3] = inFunction + 1; //1͂܂
	aPacktes[4] = inPower;
	aPacktes[5] = generateCRC(aPacktes, 5);

	return exchangeMessage(aPacktes, TIME_REPLY);
}


bool DSGatewayLib::SetLocoDirection(uint16_t address, unsigned char inDirection)
{
	unsigned char aPacktes[SIZE_PACKET];

	clearMessage(aPacktes);

	aPacktes[0] = CMD_DIRECTION | 0b0101;
	aPacktes[1] = lowbyte(address);
	aPacktes[2] = highbyte(address);
	aPacktes[3] = inDirection - 1;// FWD 1->0, REV:2->1
	aPacktes[4] = generateCRC(aPacktes, 4);

	return exchangeMessage(aPacktes, TIME_REPLY);
}

/*
bool DSGatewayLib::SetTurnout(uint16_t address, bool straight)
{
	uint8_t aSwitch = straight == true ? (uint8_t)1 : (uint8_t)0;
	
	return SetTurnout(address, aSwitch);
}*/

bool DSGatewayLib::SetTurnout(uint16_t address, uint8_t inSwitch)
{
	unsigned char aPacktes[SIZE_PACKET];

	clearMessage(aPacktes);

	aPacktes[0] = CMD_ACCESSORY | 0b0110;
	aPacktes[1] = lowbyte(address);
	aPacktes[2] = highbyte(address);
	aPacktes[3] = 0x00; // position
	aPacktes[4] = convertAcc_MMDCC(address, inSwitch);// 0: Straight, 1: diverging from DS 1:straight, 0: diverging
	aPacktes[5] = generateCRC(aPacktes, 4);

	return exchangeMessage(aPacktes, TIME_REPLY);
}

uint8_t DSGatewayLib::convertAcc_MMDCC(uint16_t address, uint8_t inSwitch)
{
	
	switch( GetLocIDProtocol(address >> 8))
	{
	case ADDR_ACC_MM2:
		/* 0:Straight, 1: diverging */
		return (inSwitch == 0) ? 1 : 0;
		break;
		
	case ADDR_ACC_DCC:
		/* 1:Straight, 0: diverging */
		return inSwitch;
		break;
		
	default:
		return inSwitch;
		break;
	}
	
}

uint16_t DSGatewayLib::GetLocIDProtocol(uint8_t address)
{
	if( address < 0x04)
	{
			return ADDR_MM2;
	}
	else if( (address >= 0x30) && (address <= 0x33))
	{
			return ADDR_ACC_MM2;
	}
	else if( (address >= 0x38) && (address <= 0x3F))
	{
			return ADDR_ACC_DCC;
	}
	else if( (address >= 0x40) && (address <= 0x70))
	{
			return ADDR_MFX;
	}
	else if( (address >= 0xC0) && (address <= 0xFF))
	{
			return ADDR_DCC;
	}
	else
	{
		return 0;
	}
}

bool DSGatewayLib::WriteConfig(uint16_t address, uint16_t number, uint8_t value)
{
	unsigned char aPacktes[SIZE_PACKET];
	
	uint16_t aOffsetCVNo;
	
	if( address >= ADDR_DCC)
	{
		aOffsetCVNo = number;
	}
	else
	{
		aOffsetCVNo = number | 0x8000;
	}

	clearMessage(aPacktes);

	aPacktes[0] = CMD_CVWRITE | 0b0101;
	aPacktes[1] = lowbyte(aOffsetCVNo);
	aPacktes[2] = highbyte(aOffsetCVNo);
	aPacktes[3] = value; 
	aPacktes[4] = generateCRC(aPacktes, 3);

	return exchangeMessage(aPacktes, TIME_REPLY);
}

bool DSGatewayLib::ReadConfig(uint16_t address, uint16_t number, uint8_t *value)
{
	unsigned char aPacktes[SIZE_PACKET];

	clearMessage(aPacktes);

	aPacktes[0] = CMD_CVREAD | 0b0100;
	aPacktes[1] = lowbyte(number);
	aPacktes[2] = highbyte(number);
	aPacktes[3] = generateCRC(aPacktes, 2);
	
	*value = 0;

	return exchangeMessage(aPacktes, TIME_REPLY);
}

unsigned char DSGatewayLib::reverseByte(unsigned char inByte)
{
	
	unsigned char aTemp = 0;
	
	for( int i = 0; i < 8; i++)
	{
		aTemp = aTemp | (((inByte >> i) & 0b1) << (7 - i));
	}	
	
	return aTemp;
	
}

unsigned char DSGatewayLib::generateCRC(unsigned char *inPackets, unsigned char inLen)
{
	unsigned char aCRC = inPackets[0];

	for( int i = 1; i < inLen; i++)
	{
		aCRC = aCRC ^ inPackets[i];
	}
	
	return aCRC;

}

uint8_t DSGatewayLib::lowbyte(uint16_t inWordData)
{
	return (uint8_t)(inWordData & 0xFF);
}

uint8_t DSGatewayLib::highbyte(uint16_t inWordData)
{
	return (uint8_t)((inWordData >> 8) & 0xFF);
}


