/*********************************************************************
 * Desktop Station Gateway Library for Raspberry Pi
 *
 * Copyright (C) 2015 Yaasan
 *
 */

#ifndef DSGATEWAYLIB_H
#define DSGATEWAYLIB_H

#include <stdint.h>
#include <stdio.h>
#include <wiringPi.h>
#include <wiringPiSPI.h>
#include <unistd.h>


#define DSGATEWAY_VERSION 0x01
#define DEBUG false
#define SIZE_PACKET 8
#define TIME_REPLY 200
#define SPI_SSPIN 10
#define SPI_CHANNEL 0
#define SPI_SPEED 1000000

#define ADDR_MM2     0x0000 // MM2 locomotive
#define ADDR_SX1     0x0800 // Selectrix (old) locomotive
#define ADDR_MFX     0x4000 // MFX locomotive
#define ADDR_SX2     0x8000 // Selectrix (new) locomotive
#define ADDR_DCC     0xC000 // DCC locomotive
#define ADDR_ACC_SX1 0x2000 // Selectrix (old) magnetic accessory
#define ADDR_ACC_MM2 0x2FFF // MM2 magnetic accessory
#define ADDR_ACC_DCC 0x37FF // DCC magnetic accessory

#define SPEEDSTEP_DCC28 0
#define SPEEDSTEP_DCC14 1
#define SPEEDSTEP_DCC127 2
#define SPEEDSTEP_MM14 0
#define SPEEDSTEP_MM28 1

#define CMD_PWR_OFF 0b00000000
#define CMD_PWR_ON  0b11110000

#define CMD_WAIT 	0b11010000
#define CMD_OK		0b10000000
#define CMD_CRCERR	0b10010000
#define CMD_CMDERR	0b10100000
#define CMD_UNKERR	0b11000000
#define CMD_DCC_IDLE	0b11000000
#define CMD_SPEED		0b00010000
#define CMD_ACCESSORY	0b00100000
#define CMD_FUNCTION	0b00110000
#define CMD_CVWRITE		0b01000000
#define CMD_DIRECTION	0b01010000
#define CMD_CVREAD		0b01100000
#define CMD_EXTENTION	0b01110000


/* Eg@\p */
#define EXCMD_NONE		0


/* Defined functions */
class DSGatewayLib
{
  private:
	void sendPacket(unsigned char *inPackets);
	bool sendMessage(unsigned char *inPackets);
	bool exchangeMessage(unsigned char *inPackets, uint16_t timeout);
	void clearMessage(unsigned char *inPackets);
	unsigned char generateCRC(unsigned char *inPackets, unsigned char inLen);
	uint8_t convertAcc_MMDCC(uint16_t address, uint8_t inSwitch);
	bool poweron;
	unsigned char reverseByte(unsigned char inByte);
	uint8_t lowbyte(uint16_t inWordData);
	uint8_t highbyte(uint16_t inWordData);
	
  public:
	DSGatewayLib();
	~DSGatewayLib();
	void begin();
	
	bool IsPower();
	
	bool SetPower(bool power);
	bool SetPowerEx(bool power);
	bool SetLocoSpeed(uint16_t address, int inSpeed);
	bool SetLocoSpeedEx(uint16_t address, int inSpeed, int inProtcol);
	bool SetLocoFunction(uint16_t address, unsigned char inFunction, unsigned char inPower);
	bool SetLocoDirection(uint16_t address, unsigned char inDirection);
	//bool SetTurnout(uint16_t address, bool straight);
	bool SetTurnout(uint16_t address, uint8_t inSwitch);
	bool WriteConfig(uint16_t address, uint16_t number, uint8_t value);
	bool ReadConfig(uint16_t address, uint16_t number, uint8_t *value);
	uint16_t GetLocIDProtocol(uint8_t address);

};

#endif
